/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { TypedEventEmitter } from "@fluid-internal/client-utils";
import { summarizerClientType } from "./summary/index.js";
/**
 * Proxy to the real IDeltaManager for restricting certain access to layers below container runtime in summarizer clients:
 * - Summarizer client should be read-only to layers below the container runtime to restrict local changes.
 * - Summarizer client should not be active to layers below the container runtime to restrict local changes.
 */
export class DeltaManagerSummarizerProxy extends TypedEventEmitter {
    get IDeltaSender() {
        return this;
    }
    get inbound() {
        return this.deltaManager.inbound;
    }
    get outbound() {
        return this.deltaManager.outbound;
    }
    get inboundSignal() {
        return this.deltaManager.inboundSignal;
    }
    get minimumSequenceNumber() {
        return this.deltaManager.minimumSequenceNumber;
    }
    get lastSequenceNumber() {
        return this.deltaManager.lastSequenceNumber;
    }
    get lastMessage() {
        return this.deltaManager.lastMessage;
    }
    get lastKnownSeqNumber() {
        return this.deltaManager.lastKnownSeqNumber;
    }
    get initialSequenceNumber() {
        return this.deltaManager.initialSequenceNumber;
    }
    get hasCheckpointSequenceNumber() {
        return this.deltaManager.hasCheckpointSequenceNumber;
    }
    get clientDetails() {
        return this.deltaManager.clientDetails;
    }
    get version() {
        return this.deltaManager.version;
    }
    get maxMessageSize() {
        return this.deltaManager.maxMessageSize;
    }
    get serviceConfiguration() {
        return this.deltaManager.serviceConfiguration;
    }
    get active() {
        // Summarize clients should not be active. There shouldn't be any local changes (writes) in the summarizer
        // except for the SummarizeOp which is generated by the runtime.
        return !this.isSummarizerClient && this.deltaManager.active;
    }
    get readOnlyInfo() {
        // Summarizer clients should be read-only as far as the runtime and layers below are concerned. There shouldn't
        // be any local changes (writes) in the summarizer except for the summarize op which is generated by the runtime.
        if (this.isSummarizerClient) {
            return {
                readonly: true,
                forced: false,
                permissions: undefined,
                storageOnly: false,
            };
        }
        return this.deltaManager.readOnlyInfo;
    }
    constructor(deltaManager) {
        super();
        this.deltaManager = deltaManager;
        this.onPrepareSend = (messageBuffer) => {
            this.emit("prepareSend", messageBuffer);
        };
        this.onSubmitOp = (message) => {
            this.emit("submitOp", message);
        };
        this.onOp = (message, processingTime) => {
            this.emit("op", message, processingTime);
        };
        this.onPong = (latency) => {
            this.emit("pong", latency);
        };
        this.onConnect = (details, opsBehind) => {
            this.emit("connect", details, opsBehind);
        };
        this.onDisconnect = (reason, error) => {
            this.emit("disconnect", reason, error);
        };
        this.onReadonly = (readonly, readonlyConnectionReason) => {
            this.emit("readonly", readonly, readonlyConnectionReason);
        };
        // We are expecting this class to have many listeners, so we suppress noisy "MaxListenersExceededWarning" logging.
        super.setMaxListeners(0);
        this.isSummarizerClient = this.deltaManager.clientDetails.type === summarizerClientType;
        this.deltaManager.on("prepareSend", this.onPrepareSend);
        this.deltaManager.on("submitOp", this.onSubmitOp);
        this.deltaManager.on("op", this.onOp);
        this.deltaManager.on("pong", this.onPong);
        this.deltaManager.on("connect", this.onConnect);
        this.deltaManager.on("disconnect", this.onDisconnect);
        this.deltaManager.on("readonly", this.onReadonly);
    }
    dispose() {
        this.deltaManager.off("prepareSend", this.onPrepareSend);
        this.deltaManager.off("submitOp", this.onSubmitOp);
        this.deltaManager.off("op", this.onOp);
        this.deltaManager.off("pong", this.onPong);
        this.deltaManager.off("connect", this.onConnect);
        this.deltaManager.off("disconnect", this.onDisconnect);
        this.deltaManager.off("readonly", this.onReadonly);
    }
    submitSignal(content, targetClientId) {
        return this.deltaManager.submitSignal(content, targetClientId);
    }
    flush() {
        return this.deltaManager.flush();
    }
}
//# sourceMappingURL=deltaManagerSummarizerProxy.js.map